# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/grid/connections/http_connection.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n,proto/grid/connections/http_connection.proto\x12\x15syft.grid.connections"Q\n\x12GridHTTPConnection\x12\x10\n\x08\x62\x61se_url\x18\x01 \x01(\t\x12\x15\n\rsession_token\x18\x02 \x01(\t\x12\x12\n\ntoken_type\x18\x03 \x01(\tb\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "proto.grid.connections.http_connection_pb2", globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _GRIDHTTPCONNECTION._serialized_start = 71
    _GRIDHTTPCONNECTION._serialized_end = 152
# @@protoc_insertion_point(module_scope)
